/*-----------------------------------------------------------------------

File  : cco_proofproc.c

Author: Stephan Schulz

Contents
 
  Functions realizing the proof procedure.

  Copyright 1998, 1999 by the author.
  This code is released under the GNU General Public Licence.
  See the file COPYING in the main CLIB directory for details.
  Run "eprover -h" for contact information.

Changes

<1> Mon Jun  8 11:47:44 MET DST 1998
    New

-----------------------------------------------------------------------*/

#include "cco_proofproc.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: remove_unit_subsumed()
//
//   Remove all clauses subsumed by subsumer from set, kill their
//   children. Return number of removed clauses.
//
// Global Variables: -
//
// Side Effects    : Changes set, memory operations.
//
/----------------------------------------------------------------------*/

static long remove_unit_subsumed(Clause_p subsumer, ClauseSet_p set)
{
   Clause_p handle, next;
   long     res = 0;

   assert(ClauseLiteralNumber(subsumer) == 1);

   DEBUGMARK(PP_HIGHDETAILS , "remove_unit_subsumed()...\n");
   next = set->anchor->succ;
   while((handle = ClauseSetFindUnitSubsumedClause(set, next,
						   subsumer)))
   {
      res++;
      DEBUGMARK(PP_HIGHDETAILS, "*\n");
      next = handle->succ;      
      ClauseKillChildren(handle);
      if(set->demod_index)
      {	 
	 ClauseSetIndexedDeleteEntry(handle);      
      }
      else
      {
	 ClauseSetDeleteEntry(handle);  
      }
   }
   DEBUGMARK(PP_HIGHDETAILS, "...remove_unit_subsumed()\n");

   return res;
}

/*-----------------------------------------------------------------------
//
// Function: remove_subsumed()
//
//   Remove all clauses subsumed by subsumer from set, kill their
//   children. Return number of removed clauses.
//
// Global Variables: -
//
// Side Effects    : Changes set, memory operations.
//
/----------------------------------------------------------------------*/

static long remove_subsumed(Clause_p subsumer, ClauseSet_p set)
{
   Clause_p handle, next;
   long     res = 0;

   assert(ClauseLiteralNumber(subsumer) > 1);

   DEBUGMARK(PP_HIGHDETAILS , "remove_subsumed()...\n");
   next = set->anchor->succ;
   while((handle = ClauseSetFindSubsumedClause(set, next,
					       subsumer)))
   {
      res++;
      DEBUGMARK(PP_HIGHDETAILS, "*\n");
      next = handle->succ;      
      ClauseKillChildren(handle);
      if(set->demod_index)
      {	 
	 ClauseSetIndexedDeleteEntry(handle);      
      }
      else
      {
	 ClauseSetDeleteEntry(handle);  
      }
   }
   DEBUGMARK(PP_HIGHDETAILS, "...remove_subsumed()\n");

   return res;
}


/*-----------------------------------------------------------------------
//
// Function: eleminate_backward_rewritten_clauses()
//
//   Remove all processed clauses rewritable with clause and put
//   them into state->tmp_store.
//
// Global Variables: -
//
// Side Effects    : Changes clause sets
//
/----------------------------------------------------------------------*/

static SysDate
eleminate_backward_rewritten_clauses(ProofState_p
				     state, ProofControl_p control,
				     Clause_p clause)
{
   SysDate date = ClauseSetListGetMaxDate(state->demods,
					  FullRewrite);
   long res = 0, tmp;

   if(ClauseIsDemodulator(clause))
   {
      SysDateInc(&date);
      tmp = RemoveClausesWithRewritableMaxSides(control->ocb,
						state->processed_pos_rules,
						state->tmp_store,
						clause, date,
						&(state->clause_count));
      /* printf("Removed %ld rules by bw-rewriting.\n", tmp); */
      res += tmp;
      tmp = RemoveClausesWithRewritableMaxSides(control->ocb,
						state->processed_pos_eqns,
						state->tmp_store,
						clause, date,
						&(state->clause_count));
      /* printf("Removed %ld equations by bw-rewriting.\n", tmp); */
      res += tmp;
      tmp = RemoveClausesWithRewritableMaxSides(control->ocb, 
						state->processed_neg_units,
						state->tmp_store,
						clause, date,
						&(state->clause_count));
      /* printf("Removed %ld neg unit clauses by bw-rewriting.\n", tmp); */
      res += tmp;
      res += RemoveClausesWithRewritableMaxSides(control->ocb, 
						 state->processed_non_units,
						 state->tmp_store,
						 clause, date,
						&(state->clause_count));
      state->backward_rewritten_count+=res;
   }
   /* printf("Removed %ld clauses\n", res); */
   return date;
}
   

/*-----------------------------------------------------------------------
//
// Function: eleminate_backward_subsumed_clauses()
//
//   Eleminate subsumed processed clauses, return number of clauses
//   deleted. 
//
// Global Variables: -
//
// Side Effects    : Changes clause sets.
//
/----------------------------------------------------------------------*/


static long eleminate_backward_subsumed_clauses(ProofState_p state,
						Clause_p clause)
{
   long res = 0;

   if(ClauseLiteralNumber(clause) == 1)
   {
      if(clause->pos_lit_no)
      {
	 /* res += remove_subsumed(clause,
	    state->processed_pos_rules); Should be impossible! */
	 res += remove_unit_subsumed(clause, state->processed_pos_eqns);
	 res += remove_unit_subsumed(clause, state->processed_non_units);
      }
      else
      {
	 res += remove_unit_subsumed(clause, state->processed_neg_units);
	 res += remove_unit_subsumed(clause, state->processed_non_units);
      }
   }
   else
   {
      res += remove_subsumed(clause, state->processed_non_units);
   }
   state->backward_subsumed_count+=res;
   return res;
}

/*-----------------------------------------------------------------------
//
// Function: interreduce_clause_sets()
//
//   Rewrite all clause sets in state with
//   state->processed_pos_clauses. Assumes that all maximal sides of
//   maximal literals are irreducible!
//
// Global Variables: -
//
// Side Effects    : Changes clauses in sets.
//
/----------------------------------------------------------------------*/

static void interreduce_clause_sets(ProofState_p state, ProofControl_p
				    control)
{
   bool res = false, tmp;

   tmp = ClauseSetComputeLINormalform(control->ocb,
				      state->terms,
				      state->processed_pos_rules,
				      state->demods,
				      FullRewrite, res);
   res = res || tmp;
   /* These clauses can never be rewritten, as they only have maximal
      terms 
     tmp = ClauseSetComputeLINormalform(control->ocb,
				      state->terms,
				      state->processed_pos_eqns,
				      state->demods, FullRewrite, res);
				      res = res || tmp; 
   */
   ClauseSetComputeLINormalform(control->ocb,
				state->terms,
				state->processed_non_units,
				state->demods, FullRewrite, res);
   res = res || tmp;   
   tmp = ClauseSetComputeLINormalform(control->ocb,
				state->terms,
				state->processed_neg_units,
				state->demods, FullRewrite, res);
}



/*-----------------------------------------------------------------------
//
// Function: insert_new_clauses()
//
//   Rewrite clauses in state->tmp_store, remove superfluous literals,
//   insert them into state->unprocessed. If an empty clause is
//   detected, return it, otherwise return NULL.
//
// Global Variables: -
//
// Side Effects    : As described.
//
/----------------------------------------------------------------------*/

Clause_p insert_new_clauses(ProofState_p state, ProofControl_p control)
{
   Clause_p handle;
   int      removed_lits;

   state->generated_count+=state->tmp_store->members;	 
   while((handle = ClauseSetExtractFirst(state->tmp_store)))
   {
      DEBUGMARK(PP_INSERTING_NEW, "insert_new_clauses() - processing: ");
      DEBUG(PP_INSERTING_NEW, ClausePrint(stdout, handle, true););
      DEBUG(PP_INSERTING_NEW, printf("\n"););
      ClauseComputeLINormalform(control->ocb,
				state->terms, handle,
				state->demods, control->forward_demod);
      DEBUGMARK(PP_INSERTING_NEW, "insert_new_clauses() normal form build\n");
      removed_lits = ClauseRemoveSuperfluousLiterals(handle);
      if(removed_lits)
      {
	 ClauseDefaultDocumentModification(handle, CM_MINIMIZE, NULL);
      }
      if(control->ac_handling_active&&control->ac_res_aggressive)
      {
	 removed_lits = ClauseRemoveACResolved(handle);
	 if(removed_lits)
	 {
	    ClauseDefaultDocumentModification(handle, CM_ACRESOLVE, NULL);
	 }
      }
      if(ClauseIsEmpty(handle))
      {
	 return handle;	 
      }
      DEBUGMARK(PP_INSERTING_NEW, "insert_new_clauses() Literals removed\n");
      if(ClauseIsTrivial(handle))
      {
	 assert(!handle->children);
	 ClauseDetachParents(handle);
	 ClauseDefaultDocumentStatusChange(handle, CS_DELETE);
	 ClauseFree(handle);
      }
      else
      {
	 state->non_trivial_generated_count++;
	 ClauseDelProp(handle, CPIsOriented);
	 if(!control->select_on_proc_only)
	 {
	    DoLiteralSelection(control, handle); 
	 }
	 else
	 {
	    SelectNoLiterals(control->ocb,handle);
	 }
	 HCBClauseEvaluate(control->hcb, handle);
	 ClauseDefaultDocumentEvaluation(handle);
	 ClauseSetInsert(state->unprocessed, handle);
	 ClauseDelProp(handle, CPIsOriented);
      }
   }
   return NULL;
}


/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/



/*-----------------------------------------------------------------------
//
// Function: ProofControlInit()
//
//   Initialize a proof control cell for a given proof state (with at
//   least axioms and signature) and a set of parameters
//   describing the ordering and heuristics.
//
// Global Variables: -
//
// Side Effects    : Sets specs.
//
/----------------------------------------------------------------------*/

void ProofControlInit(ProofState_p state,ProofControl_p control,
		      HeuristicParms_p params)
{
   PStackPointer sp;
   Scanner_p in;

   assert(control && control->wfcbs);
   assert(state && state->axioms && state->signature);
   assert(params && params->wfcb_definitions &&
	  params->hcb_definitions);
   assert(!control->ocb);
   assert(!control->hcb);
   
   SpecFeaturesCompute(&(control->problem_specs),
		       state->axioms);

   control->ocb = TOSelectOrdering(state, params,
				   &(control->problem_specs));

   in = CreateScanner(StreamTypeInternalString,
		      DefaultWeightFunctions,
		      true, NULL);
   WeightFunDefListParse(control->wfcbs, in, control->ocb, state);
   DestroyScanner(in);
   for(sp = 0; sp < PStackGetSP(params->wfcb_definitions); sp++)
   {
      in = CreateScanner(StreamTypeOptionString,
			 PStackElementP(params->wfcb_definitions, sp) ,
			 true, NULL);
      WeightFunDefListParse(control->wfcbs, in, control->ocb, state);
      DestroyScanner(in);
   }
   in = CreateScanner(StreamTypeInternalString,
		      DefaultHeuristics,
		      true, NULL);
   HeuristicDefListParse(control->hcbs, in, control->wfcbs,
			 control->ocb, state); 
   DestroyScanner(in);
   for(sp = 0; sp < PStackGetSP(params->hcb_definitions); sp++)
   {
      in = CreateScanner(StreamTypeOptionString,
			 PStackElementP(params->hcb_definitions, sp) ,
			 true, NULL);
      HeuristicDefListParse(control->hcbs, in, control->wfcbs,
			    control->ocb, state); 
      DestroyScanner(in);
   }
   control->ac_handling = params->ac_handling;
   control->ac_res_aggressive = params->ac_res_aggressive;
   control->filter_limit = params->filter_limit;
   control->filter_copies_limit = params->filter_copies_limit;
   control->reweight_limit = params->reweight_limit;
   control->delete_bad_limit = params->delete_bad_limit;
   control->er_varlit_destructive = params->er_varlit_destructive;
   control->prefer_initial_clauses = params->prefer_initial_clauses;
   control->select_on_proc_only = params->select_on_proc_only;      
   control->inherit_paramod_lit = params->inherit_paramod_lit;      
   control->selection_strategy = params->selection_strategy;      
   control->pos_lit_sel_min = params->pos_lit_sel_min;      
   control->pos_lit_sel_max = params->pos_lit_sel_max;      
   control->neg_lit_sel_min = params->neg_lit_sel_min;      
   control->neg_lit_sel_max = params->neg_lit_sel_max;      
   control->all_lit_sel_min = params->all_lit_sel_min;      
   control->all_lit_sel_max = params->all_lit_sel_max;      
   control->weight_sel_min  = params->weight_sel_min;      
   control->forward_demod   = params->forward_demod;
   control->hcb = GetHeuristic(params->heuristic_name,
			       state,
			       control, 
			       params);
}



/*-----------------------------------------------------------------------
//
// Function: ProofStateInit()
//
//   Given a proof state with axioms and a heuristic parameter
//   description, initialize the ProofStateCell, i.e. generate the
//   HCB, the ordering, and evaluate the axioms and put them in the
//   unprocessed list.
//
// Global Variables: 
//
// Side Effects    : 
//
/----------------------------------------------------------------------*/

void ProofStateInit(ProofState_p state, ProofControl_p control,
		    HeuristicParms_p h_parms)
{
   Clause_p handle, new;
   HCB_p    tmphcb;
   PStack_p traverse;
   Eval_p   cell;

   OUTPRINT(1, "# Initializing proof state\n");
   
   DEBUGMARK(PP_LOWDETAILS,"ProofStateInit()...starting to copy\n");
   
   while(!ClauseSetEmpty(state->processed_pos_rules))
   {
      handle = DemodExtract(state->processed_pos_rules->anchor->succ);
      HCBClauseEvaluate(control->hcb, handle);
      ClauseDefaultDocumentEvaluation(handle);
      ClauseSetInsert(state->unprocessed, handle);
   }
   while(!ClauseSetEmpty(state->processed_pos_eqns))
   {
      handle = DemodExtract(state->processed_pos_eqns->anchor->succ);
      HCBClauseEvaluate(control->hcb, handle);
      ClauseDefaultDocumentEvaluation(handle);
      ClauseSetInsert(state->unprocessed, handle);
   }

   state->processed_pos_rules->demod_index = PDTreeAlloc();
   state->processed_pos_eqns->demod_index = PDTreeAlloc();

   while(!ClauseSetEmpty(state->processed_non_units))
   {
      handle = state->processed_non_units->anchor->succ;
      ClauseSetExtractEntry(handle);
      HCBClauseEvaluate(control->hcb, handle);
      ClauseDefaultDocumentEvaluation(handle);
      ClauseSetInsert(state->unprocessed, handle);
   }
   
   tmphcb = GetHeuristic("Uniq", state, control, h_parms);
   ClauseSetReweight(tmphcb, state->axioms);
   
   traverse =
      EvalTreeTraverseInit(
	 PDArrayElementP(state->axioms->eval_indices,0));
   
   while((cell = EvalTreeTraverseNext(traverse)))
   {
      handle = cell->object;
      new = ClauseCopy(handle, state->terms);
      ClauseSetProp(new, CPInitial);
      new->ident = ++(state->clause_count);
      ClauseDefaultDocumentCreation(new, CC_AXIOM);
      HCBClauseEvaluate(control->hcb, new);
      ClauseDefaultDocumentEvaluation(new);
      if(ClauseIsGoal(new))
      {
	 ClauseSetProp(new, CPIsSOS);
      }
      else
      {
	 ClauseDelProp(new, CPIsSOS);
      }
      if(control->prefer_initial_clauses)
      {
	 EvalListChangePriority(new->evaluations, PrioDefer);
      }
      ClauseSetInsert(state->unprocessed, new);
   }
   EvalTreeTraverseExit(traverse);
   if(control->ac_handling!=NoACHandling)
   {
      if(OutputLevel)
      {
	 fprintf(GlobalOut, "# Scanning for AC axioms\n");
      }
      control->ac_handling_active = ClauseSetScanAC(state->signature,
						    state->unprocessed);
      if(OutputLevel)
      {      
	 SigPrintACStatus(GlobalOut, state->signature);
	 if(control->ac_handling_active)
	 {
	    fprintf(GlobalOut, "# AC handling enabled\n");
	 }
      }
   }
}


/*-----------------------------------------------------------------------
//
// Function: ProcessClause()
//
//   Select an unprocessed clause, process it. Return pointer to empty
//   clause if it can be derived, NULL otherwise.
//
// Global Variables: -
//
// Side Effects    : Everything ;-)
//
/----------------------------------------------------------------------*/

Clause_p ProcessClause(ProofState_p state, ProofControl_p control)
{
   Clause_p clause, tmp_copy, empty;
   SysDate  clausedate;
   long     clause_count;

   DEBUGMARK(PP_LOWDETAILS, "ProcessClause...\n");
   clause = control->hcb->hcb_select(control->hcb,
					 state->unprocessed);   
   if(OutputLevel==1)
   {
      putc('#', GlobalOut);
   }
   /*
   printf("\n");
   EvalListPrintComment(stdout, 
			clause->evaluations);printf("# ");
   ClausePrint(stdout, clause, true);
   printf("\n"); */
   
   assert(clause);
   
   state->processed_count++;
   
   ClauseSetExtractEntry(clause);
   ClauseSetProp(clause, CPIsProcessed);
   ClauseDetachParents(clause);
   ClauseRemoveEvaluations(clause);

   if(!ForwardContractClause(state, control, clause, true, FullRewrite))
   {
      return NULL;
   }
   if(control->ac_handling!=NoACHandling)
   {
      bool res;
      res = ClauseScanAC(state->signature, clause);
      if(res && !control->ac_handling_active)
      {
	 control->ac_handling_active = true;
	 if(OutputLevel)
	 {      
	    SigPrintACStatus(GlobalOut, state->signature);
	    fprintf(GlobalOut, "# AC handling enabled dynamically\n");
	 }
      }
   }

   ClauseDefaultDocumentProcessing(clause, CP_PROCESS);

   if(control->er_varlit_destructive &&
      (clause_count = ClauseERNormalizeVar(state->terms,
					   clause,
					   state->tmp_store,
					   state->freshvars,
					   &(state->clause_count))))
   {
      state->other_redundant_count += clause_count;
      state->resolv_count += clause_count;
      if((empty = insert_new_clauses(state, control)))
      {
	 return empty;
      }
      return NULL;
   }

   if(ClauseIsEmpty(clause))
   {
      return clause;
   }

   state->proc_non_trivial_count++;
   
   /* Now on to backward subsumption. */   

   clausedate = eleminate_backward_rewritten_clauses(state, control,
						     clause);
   
   eleminate_backward_subsumed_clauses(state, clause);
  
   tmp_copy = ClauseCopy(clause, state->tmp_terms);      
   tmp_copy->ident = clause->ident;

   ClauseMarkRestrictedTerms(clause);
   ClauseDefaultDocumentModification(clause, CM_TERM_RENAME, NULL);
   
   if(ClauseIsDemodulator(clause))
   {
      assert(clause->neg_lit_no == 0);
      clause->date = clausedate;
      if(EqnIsOriented(clause->literals))
      {
	 state->processed_pos_rules->date = clausedate;
	 DemodInsert(state->processed_pos_rules, clause);
      }
      else
      {
	 state->processed_pos_eqns->date = clausedate;
	 DemodInsert(state->processed_pos_eqns, clause);
      }
      interreduce_clause_sets(state, control);
   }
   else if(ClauseLiteralNumber(clause) == 1)
   {
      assert(clause->neg_lit_no == 1);
      DemodInsert(state->processed_neg_units, clause);
   }
   else
   {
      DemodInsert(state->processed_non_units, clause);
   }
      
   if(control->selection_strategy != SelectNoGeneration)
   {
      state->factor_count+=
	 ComputeAllEqualityFactors(state->terms, control->ocb, clause,
				   state->tmp_store, state->freshvars,
				   &(state->clause_count));
      state->resolv_count+=
	 ComputeAllEqnResolvents(state->terms, clause, state->tmp_store,
				 state->freshvars, &(state->clause_count));
      state->paramod_count+=
	 ComputeAllParamodulants(state->terms, control->ocb,
				 tmp_copy, clause,
				 state->processed_pos_rules,
				 state->tmp_store, state->freshvars,
				 &(state->clause_count)); 
      state->paramod_count+=
	 ComputeAllParamodulants(state->terms, control->ocb,
				 tmp_copy, clause,
				 state->processed_pos_eqns,
				 state->tmp_store, state->freshvars,
				 &(state->clause_count)); 
      state->paramod_count+=
	 ComputeAllParamodulants(state->terms, control->ocb,
				 tmp_copy, clause,
				 state->processed_neg_units,
				 state->tmp_store, state->freshvars,
				 &(state->clause_count));
      state->paramod_count+=
	 ComputeAllParamodulants(state->terms, control->ocb,
				 tmp_copy, clause,
				 state->processed_non_units, 
				 state->tmp_store, state->freshvars,
				 &(state->clause_count));
   }
   ClauseFree(tmp_copy);      
   
   if((empty = insert_new_clauses(state, control)))
   {
      return empty;
   }  
   return NULL;
}


/*-----------------------------------------------------------------------
//
// Function:  Saturate()
//
//   Process clauses until either the empty clause has been derived, a
//   specified number of clauses has been processed, or the clause set
//   is saturated. Return emoty clause (if found) or NULL. 
//
// Global Variables: -
//
// Side Effects    : Modifies state.
//
/----------------------------------------------------------------------*/

Clause_p Saturate(ProofState_p state, ProofControl_p control, long
		  step_limit, long proc_limit, long unproc_limit, long
		  total_limit)
{
   Clause_p unsatisfiable = NULL;
   long count = 0;
   long filter_base   = ProofStateStorage(state);
   long reweight_base = state->unprocessed->members;
   long filter_copies_base = ProofStateStorage(state);
   long tmp;

   while(!TimeIsUp &&
	 state->unprocessed->members && 
	 step_limit>count && 
	 proc_limit>(state->processed_pos_rules->members +
		     state->processed_pos_eqns->members +
		     state->processed_neg_units->members +
		     state->processed_non_units->members) &&
	 unproc_limit>state->unprocessed->members &&
	 total_limit>(state->processed_pos_rules->members +
		      state->processed_pos_eqns->members +
		      state->processed_neg_units->members +
		      state->processed_non_units->members+
		      state->unprocessed->members))
   {     
      DEBUGMARK(PP_LOWDETAILS,"Saturate loop entry.\n");

      count++;
      unsatisfiable = ProcessClause(state, control);
      if(unsatisfiable)
      {
	 break;
      }

      if((ProofStateStorage(state) - filter_copies_base)
	 > control->filter_copies_limit)
      {
	 tmp = ClauseSetDeleteCopies(state->unprocessed);
	 if(OutputLevel)
	 {
	    fprintf(GlobalOut, 
		    "# Deleted %ld clause copies (remaining: %ld)\n",
		    tmp, state->unprocessed->members);
	 }
	 state->other_redundant_count += tmp;
	 filter_copies_base = ProofStateStorage(state);
      }
      if((ProofStateStorage(state) - filter_base)
	 > control->filter_limit)
      {
	 tmp = state->unprocessed->members;
	 unsatisfiable = 
	    ForwardContractSet(state, control,
			       state->unprocessed, false, FullRewrite,
			       &(state->other_redundant_count));
	 
	 if(OutputLevel)
	 {
	    fprintf(GlobalOut, 
		    "# Special forward-contraction deletes %ld clauses"
		    "(remaining: %ld) \n",
		    tmp - state->unprocessed->members,
		    state->unprocessed->members);
	 }	 
	 filter_base = ProofStateStorage(state);
	 if(unsatisfiable)
	 {
	    break;
	 }	 
      } 
      if((state->unprocessed->members - reweight_base) 
	 > control->reweight_limit)
      {
	 OUTPRINT(1, "# Reweighting unprocessed clauses...\n");
	 ClauseSetReweight(control->hcb,  state->unprocessed);
	 reweight_base = state->unprocessed->members;
      }
      if(ProofStateStorage(state) > control->delete_bad_limit)
      {
	 tmp = HCBClauseSetDeleteBadClauses(control->hcb,
					    state->unprocessed, 
					    state->unprocessed->members/2);
	 state->non_redundant_deleted += tmp;
	 if(OutputLevel)
	 {
	    fprintf(GlobalOut, 
		    "# Deleted %ld bad clauses (prover may be"
		    " incomplete now)\n", tmp);
	 }
	 state->state_is_complete = false;
	 filter_base = MIN(filter_base, ProofStateStorage(state));
	 filter_copies_base = MIN(filter_copies_base,
				  ProofStateStorage(state));
      }
   }
   return unsatisfiable;
}


/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/



